#!/usr/bin/perl -w
#
# epoc_check.pl - Tool for checking EPOC source code
#
# Copyright 2001,2002 SX Design AS <www.sxdesign.com>.
# Copyright 2002 Alfred E. Heggestad
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 2 as
#    published by the Free Software Foundation.
#

use strict;
use Cwd;


#
# my variables
#
my $version = "0.01";
my @files_h = `find . -name "*.h"`;
my @files_c = `find . -name "*.c"`;
my @files_cpp = `find . -name "*.cpp"`;
my @files_S = `find . -name "*.S"`;
my @files_mmp = `find . -name "*.mmp"`;
my @files_mph = `find . -name "*.mph"`;
my @files_bldinf = `find . -name "bld.inf"`;
my @files_pkg = `find . -name "*.pkg"`;
my $n_files = scalar @files_h + scalar @files_c + scalar @files_cpp
  + scalar @files_S + scalar @files_mmp + scalar @files_mph
  + scalar @files_bldinf + scalar @files_pkg;
my $root = &cwd;
my $errors = 0;
my $verbose = 0;
my $do_all = 0;
my $do_para = 0;
my $be_quiet = 0;


#
# These words should not be in the source code
#
my @illegal_keywords = (
			"illegalword"
		       );


#
# print an error message and increase error count
#
sub error
  {
    printf(@_);
    $errors++;
  }


#
# print if verbose flag is turned on
#
sub verbose_print
  {
    return if (!$verbose);
    printf(@_);
  }


#
# print statistics
#
sub print_stats
  {
    print "Statistics:\n";
    print "~~~~~~~~~~~\n";
    print "Number of .c files processed:       " . scalar @files_c . "\n";
    print "Number of .cpp files processed:     " . scalar @files_cpp . "\n";
    print "Number of .h files processed:       " . scalar @files_h . "\n";
    print "Number of .S files processed:       " . scalar @files_S . "\n";
    print "Number of .mmp files processed:     " . scalar @files_mmp . "\n";
    print "Number of .mph files processed:     " . scalar @files_mph . "\n";
    print "Number of bld.inf files processed:  " . scalar @files_bldinf . "\n";
    print "Number of .pkg files processed:     " . scalar @files_pkg . "\n";
    print "\n";
    print "Total Number of files processed:    " . $n_files . "\n";
    print "Number of lines with errors:        $errors\n";
    print "\n";
  }


#
# check for strange white space
#
sub check_whitespace
  {
    my $file = shift;
    my $line = 0;
    my $empty_lines_count = 0;
    my $empty_lines_last = 0;

    open F, $file or warn "could not open $file ($!)";
    while(<F>) {
      chomp $_;
      my $line_len = length($_);
      $line++;
      if(m/ $/) {
	error("$file:$line has trailing space(s)\n");
      }
      elsif(m/\t$/) {
	error("$file:$line has trailing tab(s)\n");
      }
      elsif(m/^ {8}/) {
	error("$file:$line starts with 8 spaces, use tab instead\n");
      }
      elsif(m/^ {4}/) {
	error("$file:$line starts with 4 spaces, use tab instead\n");
      }

      # check for empty lines count
      if($line_len == 0) {
	$empty_lines_count++;
      } else {

	if($empty_lines_count > 0) {
	  $empty_lines_last = $empty_lines_count;
	  $empty_lines_count = 0;

	  if($empty_lines_last > 2) {
	    error("$file:$line should have maximum two empty lines ($empty_lines_last)\n");
	  }
	}
      }

    } # while

    close F;
  }


#
# check max line length and number of lines
#
# param 1 - Maximum chars per line
# param 2 - Maximum lines per file
# param 3 - Filename
#
sub check_xy_max
  {
    my $max_x = shift;
    my $max_y = shift;
    my $file = shift;
    my $line = 0;

    verbose_print("check_xy_max: $file\n");

    open F, $file or warn "could not open $file ($!)";
    while(<F>) {
      my $line_len = length($_);
      $line++;

      # check for line length
      if($line_len > $max_x) {
	error("$file:$line line is too wide (" . $line_len . " - max $max_x)\n");
      }

    }

    if($line > $max_y) {
      error("$file is too big ($line lines - max $max_y)\n");
    }
    close F;
}


#
# check that header files dont include other header files
#
sub check_header_inc_header
  {
    my $file = shift;
    my $line = 0;

    verbose_print("check_header_inc_header: $file\n");

    open F, $file or warn "could not open $file ($!)";
    while(<F>) {
      $line++;
      if(m/\#include (.*)/) {
	error("$file:$line includes $1\n");
      }
    }
    close F;
  }


my @HINTS = (
	     "Copyright",
	     "This program is free software",
	     "you can redistribute it",
             "General Public License version 2",
	     "Free Software Foundation"
	    );

#
# check that copyright header is there
#
sub check_copyright_header {
  my $file = shift;
  my $probability = 0;
  my $rec_sep = $/;

  # no \n delimiters
  undef $/; # alias $INPUT_RECORD_SEPARATOR, $RS

  open F, $file or warn "could not open $file ($!)";
  $_ = <F>;

  foreach my $hint (@HINTS) {
    $probability++ if /$hint/;
  }

  if($probability < 1) {
    error("$file:1 is missing copyright header (prob=$probability hints=" . scalar @HINTS . ")\n");
  }
  close F;

  $/ = $rec_sep;
}


#
# check for illegal words in the code
#
sub check_illegal_words {
  my $file = shift;
  my $line = 0;

  open F, $file or warn "could not open $file ($!)";
  while(<F>) {
    $line++;

    foreach my $word (@illegal_keywords) {
      if(/\W$word\W/) {
	error("$file:$line contains illegal word ($word)\n");
      }
    }
  }
  close F;
}


#
# check that hexadecimal numbers are lowercase
#
sub check_hex_lowercase {
    my $file = shift;
    my $line = 0;
    open F, $file or warn "could not open $file ($!)";
    while(<F>) {
      $line++;
      if(m/0x([0-9A-Fa-f]+)/) {
		$_ = $1;
		if(m/[A-F]+/) {
		  error("$file:$line 0x$_ should be lowercase\n");
		}
      }
    }
    close F;
  }


#
# check doxygen comments
#
sub check_doxygen {
    my $file = shift;
    my $line = 0;
    open F, $file or warn "could not open $file ($!)";
    while(<F>) {
      $line++;

      # check @param
      if(m/\@param (\w+)\s+(\w{1})(.*)/) {
	my $param = $1;
	my $first = $2;
	my $comment = "$2$3";

	if(!m/ \@param/) {
	  error("$file:$line \@param should look like this: ' * \@param'\n");
	}

	if($first =~ m/[a-z]{1}/) {
	  error("$file:$line \@param should start with a capital letter '$comment'\n");
	}

	if(!m/\.$/) {
	  error("$file:$line \@param should end with a dot '$comment'\n");
	}
      }

      # check ///<
      if(m/\/\/\/\<\s+(\w{1})(.*)/) {
	my $first = $1;
	my $comment = "$1$2";

	if($first =~ m/[a-z]{1}/) {
	  error("$file:$line Doxygen comments should start with a capital letter '$comment'\n");
	}

	if(!m/\.$/) {
	  error("$file:$line Doxygen comments should end with a dot '$comment'\n");
	}
      }



    }
    close F;
  }


#
# check for correct brackets usage
#
sub check_brackets {
  my $file = shift;
  my $line = 0;

  my @operators  = ("do", "if","for","while");
  
  open F, $file or warn "could not open $file ($!)";
  while(<F>) {
    $line++;

    if(/\W*(\w+\W*)\(/) {
      my $keyword = $1;

      next if($keyword =~ m/\(/);

      # check all the operators
      foreach my $op (@operators) {
        if($keyword =~ m/^$op\W*$/) {

	  # check for white space
          if($keyword =~ m/\W+/) {
            error("$file:$line whitespace before bracket '$keyword()'\n");
          }
        }
      }
    }
  }
  close F;
}


#
# check for many nested derefs
#
sub check_nested_derefs {
  my $file = shift;
  my $line = 0;

  open F, $file or warn "could not open $file ($!)";
  while(<F>) {
    $line++;

    if(m/(\w+\->\w+\->\w+\->\w+)/) {
	  error("$file:$line Many levels of dereferencing '$1'\n");
    }
  }
  close F;
}


#
# check MMP file syntax
#
sub check_mmp_syntax {
  my $file = shift;
  my $line = 0;
  my $line_cont = 0;

  open F, $file or warn "could not open $file ($!)";
  while(<F>) {
    $line++;

    if(m/^([a-z]+)(.*)/i) {
      my $tag = $1;
      my $value = $2;
      $value =~ s/ //g;

#      print "$tag = $value\n";

      if($line_cont == 0 && $tag =~ /[a-z]/) {
	error("$file:$line '$tag' should be uppercase\n");
      }

      # don't be too paranoid
      next if($tag =~ /^START/);
      next if($tag =~ /SOURCE/);
      next if($tag =~ /LANG/);

      if($tag =~ /MACRO/) {
	if($value =~ /[a-z]/) {
	  error("$file:$line The macro '$value' should be uppercase\n");
	}
      }
      elsif($value =~ /[A-Z]/) {
	error("$file:$line '$value' should be lowercase\n");
      }
      $line_cont = m/\\$/ ? 1 : 0;
    }
  }
  close F;
}


#
# print usage
#
sub usage {

  printf "$0 version $version - part of the GnuPoc project\n";
  printf "\n";
  printf "Usage:\n";
  printf "\n";
  printf "  $0 [options]\n";
  printf "\n";
  printf "options:";
  printf "\n";
  printf "  --help     Display help\n";
  printf "  --verbose  Verbose output\n";
  printf "  --all      All options enabled\n";
  printf "  --para     Paranoid. Check everything\n";
  printf "  --quiet    Only print warnings\n";
  exit;
}


#
# Main program
#
if (@ARGV == 0) {
  usage();
  exit();
}

while (@ARGV) {

  $_ = lc pop @ARGV;

  if(m/--help/) {
    usage();
  } elsif(m/--all/) {
    $do_all = 1;
  } elsif(m/--para/) {
    $do_para = 1;
    $do_all = 1;
  } elsif(m/--verbose/) {
    $verbose = 1;
  } elsif(m/--quiet/) {
    $be_quiet = 1;
  } else {
    printf "unknown option '$_'\n";
  }
}


#
# settings
#
my $max_chars_pr_line_c = 100;
my $max_chars_pr_line_cpp = 130;
my $max_chars_pr_line_h = 110;
my $max_chars_pr_line_S = 100;
my $max_chars_pr_line_std = 80;
my $max_chars_pr_line_pkg = 100;
my $max_lines_pr_file_c = 850;
my $max_lines_pr_file_cpp = 850;
my $max_lines_pr_file_h = 300;
my $max_lines_pr_file_S = 300;
my $max_lines_pr_file_std = 300;
my $max_lines_pr_file_pkg = 200;


#
# fix path first
#
foreach (@files_c) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_cpp) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_h) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_S) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_mmp) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_mph) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_bldinf) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}

foreach (@files_pkg) {
  s/^\.(.*)/$root$1/;
  chomp $_;
}


#
# check files
#
if ($do_all) {
  foreach my $f (@files_c) {
    check_xy_max($max_chars_pr_line_c, $max_lines_pr_file_c, $f);
    check_doxygen($f);
    check_nested_derefs($f) if $do_para;
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
    check_brackets($f);
  }

  foreach my $f (@files_cpp) {
    check_xy_max($max_chars_pr_line_cpp, $max_lines_pr_file_cpp, $f);
    check_doxygen($f);
    check_nested_derefs($f) if $do_para;
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
    check_brackets($f);
  }

  foreach my $f (@files_h) {
    check_xy_max($max_chars_pr_line_h, $max_lines_pr_file_h, $f);
    check_header_inc_header($f) if $do_para;
    check_doxygen($f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
    check_brackets($f);
  }

  foreach my $f (@files_S) {
    check_xy_max($max_chars_pr_line_S, $max_lines_pr_file_S, $f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
  }

  foreach my $f (@files_mmp) {
    check_xy_max($max_chars_pr_line_std, $max_lines_pr_file_std, $f);
    check_doxygen($f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
    check_mmp_syntax($f);
  }

  foreach my $f (@files_mph) {
    check_xy_max($max_chars_pr_line_std, $max_lines_pr_file_std, $f);
    check_doxygen($f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
    check_mmp_syntax($f);
  }

  foreach my $f (@files_bldinf) {
    check_xy_max($max_chars_pr_line_std, $max_lines_pr_file_std, $f);
    check_doxygen($f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
  }

  foreach my $f (@files_pkg) {
    check_xy_max($max_chars_pr_line_pkg, $max_lines_pr_file_pkg, $f);
    check_doxygen($f);
    check_whitespace($f);
    check_copyright_header($f);
    check_hex_lowercase($f);
    check_illegal_words($f);
  }

}

if (! $be_quiet) {
  print_stats();
}
