#! /usr/bin/perl -w
#
# sdk-unpack2.pl - Mal Minhas <malm@malm.co.uk>
#
# Part of GnuPoc project - EPOC SDK for GNU/Linux http://gnupoc.sourceforge.net/
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 2 as
#    published by the Free Software Foundation.
#

# 01/10/02 - Created by <malm@malm.co.uk> from an existing script sdk-unpack.pl
#            written by Rich Williams <rdw@perldog.com>.  
#            Notes:
#            =====
#            1) Works with i6comp02.exe (I6COMPv0.2).
#            2) Unpacks SDK into right directory structure for fixsrc.pl
#            3) Tested OK with wine 20020904 and Nokia 7650 SDK v0.9.

use strict;
use Cwd;

############################################################################

my $WINE = "/usr/bin/wine";
my $pwd = cwd();
my $I6C = "$pwd/"."i6comp02.exe"; # default is that i6comp is in script dir
my $CAB = "$pwd/"."data1.cab"; # default is that data1.cab is in script dir
my $DIR = "$pwd/"."Series60"; # default is that SDK is installed in Series60

if ($#ARGV < 0){
    &usage;
    exit;
}
if ($#ARGV == 0){
    if ($ARGV[0] eq '-h'){
      &usage;
      exit;
    }
    $DIR=$ARGV[0];
}

#
# sanity check
#
die "could not find $WINE" unless -e $WINE;
die "could not find $I6C" unless -e $I6C;
die "could not find $CAB" unless -e $CAB;

# We are only interested in installing components with <Data> in their names
install_component(get_components(), $DIR, 'Data', '<Data>');

############################################################################

sub usage {
# Print usage
    print "\n";
    print "Usage:\n";
    print "=====\n";
    print "\n";
    print "\tperl sdk-unpack2.pl [options] <installdir>\n";
    print "\n";
    print "Notes:\n";
    print "1) This script is for installing the Nokia\n";
    print "Series 60 SDK for Symbian OS.  You need to go to\n";
    print "www.forum.nokia.com to download the latest SDK.\n";
    print "2) <installdir> is relative to this script\n";
    print "and the default <installdir> is \"Series60\".\n";
    print "3) i6comp02.exe and data1.cab must both be in\n";
    print "the same directory as this script.  You should\n";
    print "get the CAB file after unzipping the SDK.\n";
    print "\n";
    print "Options:\n";
    print "  -h: print this help\n";
    print "\n";
    print "Support: mal\@malm.co.uk\n";
    print "\n"; 
}

sub get_components {
# Retrieve the components in the CAB file.
  my $result={}; 
  print " ** Getting components from $CAB... ** \n";
  open LISTCOMP, "$WINE -- $I6C c -v $CAB|" or die "Couldn't open listcomp: $!\n";
  local $/ = "\r\n"; 
  while (<LISTCOMP>) {# Input separator is \r\n so read in a line at a time
    last if /^=+/;
  }
  local $/ = "\r\n\r\n";
  while (<LISTCOMP>) {# Input separator is \r\n\r\n to pick up component records
    tr/\r//d; # MALM - replace all \r's in component record
    my $component={};
    foreach (split /\n/) {# Split each component record on \n's 
      # Keys include: Name, Status Text, File Groups, Sub Comp
      my($key, $value) = split /\s+:\s+/, $_;
      next unless $value;
      # Split value into array separated by , if Sub Comp or File Groups key
      $value = [split /,\s+/, $value] if $key =~ /(Sub Comp|File Groups)/;
      $component->{$key} = $value;
    }
    $result->{$component->{'Name'}} = $component;
    #print " COMPONENT: $component->{'Name'}\n";
  }
  close LISTCOMP;
  return $result;
}

sub install_filegroup {
# We only actually extract and install files associated with filegroups.
# We need to locate the destination to install files to using i6comp
  my $group = shift;
  
  my $targetdir;
  my $cdir = cwd();
  open LISTGROUP, "$WINE -- $I6C g -g'$group' $CAB|" or die "Couldn't open listgroup: $!\n";
  while (my $line = <LISTGROUP>) {
      if ($line =~ /Destination\s+:\s+(.*)/){
	  $targetdir=$1; 
	  $targetdir =~ tr/\r//d; # Replace all \r's in filegroup record
	  $targetdir =~ tr/\\/\//;
	  $targetdir =~ s/EPOC32/Epoc32/; # Need to ensure we merge
	  $targetdir =~ s/release/Release/; # these three directory
	  $targetdir =~ s/include/Include/; # capitalisation mismatches 
	  $targetdir =~ s/<TARGETDIR>/$DIR/;
	  $targetdir =~ s/<Shared Tools>/$DIR/;
	  print "\tInstalling \"$group\" in \"$targetdir\" with cwd of \"$cdir\"\n";
	  unless (-d $targetdir) {
	      make_dirtree($targetdir);
	  }
	  chdir($targetdir) or die "Failed to chdir ($targetdir) - $!\n";
	  # This has not been tested with all versions of Wine...
	  system("$WINE -- $I6C x -g'$group' $CAB >/dev/null");
      }
  }
  close LISTGROUP;
  chdir($cdir);
}

sub make_dirtree {
# Makes the directory tree...
  my $targetdir = shift;

  my $cdir = cwd();
  my $dirpath = "";
  my @dirarray = split (/\//, $targetdir);
  foreach my $dir (@dirarray){
      $dirpath = "$dirpath"."$dir/";
      unless (-d $dirpath){
          mkdir $dirpath or die "Failed to mkdir ($dirpath) - $!\n";
      }
  }
  chdir($cdir);
}

sub install_component {
# Go through each of the components. This function recursively 
# handles Sub Comps and File Groups.
# The Series 60 SDK has a top level <Data> component which 
# has 5 Sub Comps "SDK", "SDK Documentation", "Examples", 
# "3rd Party", "Language".  These in turn spawn their own
# Sub Comps and File Groups...
  my($comps, $basedir, $base, $name) = @_;

  return if $name =~ /Disk</;
  # We're only interested in $names with a $base (ie. Data) that
  # matches $basedir (ie. <Data>)
  return unless ((my $dir = $name) =~ s/<$base>/$basedir/);

  #$dir =~ tr/\\/\//;

  my $desc="";
  if ($comps->{$name}->{'Description'}) {
      $desc = $comps->{$name}->{'Description'};
  } 
  elsif ($comps->{$name}->{'Status Text'}) {
      $desc = $comps->{$name}->{'Status Text'};
  }
  printf "Installing \"$name\" (%s)\n", $desc;

  foreach (@{$comps->{$name}->{'Sub Comp'}}) {
    install_component($comps, $basedir, $base, $_);
  }

  foreach (@{$comps->{$name}->{'File Groups'}}) {
    install_filegroup($_);
  }
  #print "\n"; # Separate out the Component subtrees
}

############################################################################






